import grams.inputs as I
from grams.main import GRAMS
from grams.html_table_parser import fetch_tables
from grams.config import ROOT_DIR, DATA_DIR
from grams.algorithm.sm_wikidata import WikidataSemanticModelHelper
